/******************************************************************************
*
* project name:    TI-LIDS (Leckfettn-is-des-schwa)
* file name:       tilids.c
* initial date:    03/11/2000
* authors:         thomas.nussbaumer@gmx.net
*                  marcos.lopez@wol.es
*
* description:     a 4x4 puzzle game
*
*
* based on game gripple done by Doug Beeferman for the PC (1990)
*
* (parts of this source code are "recycled" from TI-Chess 3.01, too)
*
*
* compile with: tigcc -O2 tilids.c
*
* $Id: tilids.c,v 1.2 2002/11/14 15:24:35 nussbaumer Exp $
*
******************************************************************************/

#define SAVE_SCREEN
#define OPTIMIZE_ROM_CALLS
#define NO_EXIT_SUPPORT

#define USE_TI89
#define USE_TI92P
#define USE_V200

#include <tigcclib.h>


#define SHUFFLE_LOOPS 500    // number of random rotates at startup

//-----------------------------------------------------------------------------
// which keys are used for selecting sections is defined here
//
// normally these keys correspond to the numeric keypad !!
//-----------------------------------------------------------------------------
#define KEY_SECTION1  '7'    // upper left  section key
#define KEY_SECTION2  '9'    // upper right section key
#define KEY_SECTION3  '1'    // lower left  section key
#define KEY_SECTION4  '3'    // lower right section key
#define KEY_SECTION5  '5'    // center      section key


#define LIDS_VERSION "v1.10"

//-----------------------------------------------------------------------------
// some globals
//-----------------------------------------------------------------------------

// the initial setup of the board
short initial_setup[] = {1,1,3,3,1,1,3,3,2,2,4,4,2,2,4,4};

// the board itself
short board[4][4] = {};

// markers used to mark squares which must be redrawn
short redraw[4][4] = {};

// helper variable for clipping
SCR_RECT fullscreen = {{0,0,239,127}};

// offsets (mainly used for TI92plus)
short offsetx = 0;
short offsety = 0;


//-----------------------------------------------------------------------------
// here comes the sprite data ...
//-----------------------------------------------------------------------------
#include "pieces.h"

unsigned long mask[] = {
    0x000003ff,0x000003ff,0x000003ff,
    0x000003ff,0x000003ff,0x000003ff,
    0x000003ff,0x000003ff,0x000003ff,
    0x000003ff,0x000003ff,0x000003ff,
    0x000003ff,0x000003ff,0x000003ff,
    0x000003ff,0x000003ff,0x000003ff,
    0x000003ff,0x000003ff,0x000003ff,
    0x000003ff};


/*---------------------------------------------------------------------------*/
/* some defines for drawing functions                                        */
/*---------------------------------------------------------------------------*/
#define COLOR_WHITE      0
#define COLOR_LIGHTGRAY  1
#define COLOR_DARKGRAY   2
#define COLOR_BLACK      3
#define LIGHT_PLANE      0
#define DARK_PLANE       1
#define RECT_EMPTY       0
#define RECT_FILLED      1
#define A_CENTERED       0x40
#define A_SHADOWED       0x80


INT_HANDLER oldint5 = NULL;  // will hold the address of the previous handler later

volatile short mseconds50 = 0; // counts in 50ms steps (not exact 50ms, but 1000/19 ms)
volatile short seconds    = 0; // counts seconds
volatile short minutes    = 0; // counts minutes
volatile short hours      = 0; // counts hours
volatile short resettime  = 1; // used to signal a reset clock
volatile short running    = 1;
volatile short outputtime = 0;

void DrawString(short x,short y,char* s,short font,short attr);

//=============================================================================
// outputs the time to the statusline
//=============================================================================
void PrintTime(void) {
    char buffer[16];
    sprintf(buffer,"%02d:%02d:%02d", hours, minutes, seconds);
    DrawString(110,87,buffer,F_4x6,A_REPLACE);
}


//=============================================================================
// our own interrupt handler hooked on interrupt 5 (updates clock variables)
//=============================================================================
DEFINE_INT_HANDLER (myint5handler) {
    if (resettime) {
        mseconds50=seconds=minutes=hours=resettime=0;
    }
    else if (running) {
        mseconds50++;
        if (mseconds50 == 19) {
            mseconds50 = 0;
            seconds++;

            //-------------------------------------------------------
            // we have counted a second, let's update the time output
            // (done within the GetUserInput() routine)
            //-------------------------------------------------------
            outputtime=1;

            if (seconds == 60) {
                seconds = 0;
                minutes++;
                if (minutes == 60) {
                   minutes = 0;
                   hours++;
                   if (hours == 24) hours = 0;
                }
            }
        }
    }
    //---------------------------------------------------------
    // execute the old interrupt handler at the end of our own
    // to make sure that other timers which are generated using
    // the AMS timer routines like OSRegisterTimer() will still
    // work correctly
    //---------------------------------------------------------
    ExecuteHandler(oldint5);
}


/*===========================================================================*/
/* utility funcction to draw a string to both planes                         */
/* NOTE: there exists to new styles which are handled as flags               */
/*       A_CENTERED ... the given font will be used and the string will be   */
/*                      drawn centered on the screen                         */
/*       A_SHADOWED ... if this flag is set in the attributes a lightgray    */
/*                      shadow will be drawn with +1 pixel offset in all     */
/*                      two directions                                       */
/*===========================================================================*/
void DrawString(short x,short y,char* s,short font,short attr) {
    short do_shadow = attr & A_SHADOWED;

    x += offsetx;
    y += offsety;

    FontSetSys(font);

    if (attr & A_CENTERED) {
        x = (LCD_WIDTH-DrawStrWidth(s,font))/2;
        attr &= ~A_CENTERED;
    }

    attr &= ~A_SHADOWED;

    SetPlane(DARK_PLANE);
    DrawStr(x,y,s,attr);
    SetPlane(LIGHT_PLANE);
    DrawStr(x,y,s,attr);

    if (do_shadow) DrawStr(x+1,y+1,s,A_NORMAL);
}


/*===========================================================================*/
/* draw a line with given color from (x0,y0) to (x1,y1)                      */
/*===========================================================================*/
void DrawColorLine(short x0,short y0,short x1, short y1, short color) {
    SetPlane(LIGHT_PLANE);
    if (color == COLOR_WHITE || color == COLOR_DARKGRAY) {
        DrawLine(x0, y0, x1, y1, A_REVERSE);
    }
    else {
        DrawLine(x0, y0, x1, y1, A_NORMAL);
    }
    SetPlane(DARK_PLANE);
    if (color == COLOR_WHITE || color == COLOR_LIGHTGRAY) {
        DrawLine(x0, y0, x1, y1, A_REVERSE);
    }
    else {
        DrawLine(x0, y0, x1, y1, A_NORMAL);
    }
}


/*===========================================================================*/
/* draws a rectangle with given color                                        */
/* if fill == 1) the rectangle is filled                                     */
/*===========================================================================*/
void DrawColorRect(short x0,short y0,short x1, short y1, short color,short fill) {
    WIN_RECT wrect;
    SCR_RECT rect;
    void     (*df)(void*,void*,short);
    void*    parea;

    if (fill) {
        rect.xy.x0 = x0;
        rect.xy.y0 = y0;
        rect.xy.x1 = x1;
        rect.xy.y1 = y1;
        parea      = &rect;
        df         = (void(*)(void*,void*,short))ScrRectFill;
    }
    else {
        wrect.x0 = x0;
        wrect.y0 = y0;
        wrect.x1 = x1;
        wrect.y1 = y1;
        parea    = &wrect;
        df       = (void(*)(void*,void*,short))DrawClipRect;
    }

    SetPlane(LIGHT_PLANE);
    if (color == COLOR_WHITE || color == COLOR_DARKGRAY) {
        (*df)(parea,&fullscreen, A_REVERSE);
    }
    else {
        (*df)(parea,&fullscreen, A_NORMAL);
    }
    SetPlane(DARK_PLANE);
    if (color == COLOR_WHITE || color == COLOR_LIGHTGRAY) {
        (*df)(parea,&fullscreen, A_REVERSE);
    }
    else {
        (*df)(parea,&fullscreen, A_NORMAL);
    }
}


/*===========================================================================*/
/* draws a popup window                                                      */
/*===========================================================================*/
void DrawPopup(short x0,short y0,short x1, short y1) {
    DrawColorRect(x0,y0,x1,y1,COLOR_BLACK,RECT_EMPTY);
    x0++,y0++,x1--,y1--;
    DrawColorRect(x0,y0,x1,y1,COLOR_DARKGRAY,RECT_EMPTY);
    x0++,y0++,x1--,y1--;
    DrawColorRect(x0,y0,x1,y1,COLOR_LIGHTGRAY,RECT_EMPTY);
    x0++,y0++,x1--,y1--;
    DrawColorRect(x0,y0,x1,y1,COLOR_WHITE,RECT_FILLED);
}


//=============================================================================
// draws a single coin (21x21 pixels large)
//=============================================================================
void DrawField(short x,short y) {
    short i;

    for (i=0;i<2;i++) {
        short offset = (board[x][y]-1)*22;
        Sprite32(offsetx+1+x*23,offsety+1+y*23,22,mask,               GetPlane(i),SPRT_AND);
        Sprite32(offsetx+1+x*23,offsety+1+y*23,22,&sprites[i][offset],GetPlane(i),SPRT_OR);
    }

    redraw[x][y] = 0;
}


/*===========================================================================*/
/* get input key pressed by user                                             */
/*                                                                           */
/* replacement for ngetchx() and/or GKeyIn(), because:                       */
/*                                                                           */
/* (1) ngetchx() won't check the APD timer value                             */
/* (2) GKeyIn() destroys the display due to setting the status display       */
/*===========================================================================*/
short GetUserInput() {
    void *kbq = kbd_queue();
    short key;

    //-------------------------------------------
    // restart the APD timer now
    //-------------------------------------------
    OSTimerRestart(APD_TIMER);

    while (1) {
        //---------------------------------------
        // wait until we get a key
        // if the APD timer expires -> turn
        // calculator off
        //---------------------------------------
        while (OSdequeue(&key, kbq)) {
            if (OSTimerExpired(APD_TIMER)) {
                off();
                OSTimerRestart(APD_TIMER);
            }
            if (outputtime) {
                outputtime = 0;
                PrintTime();
            }
        }

        //---------------------------------------
        // handle calculator-off key combinations
        // if the APD timer expires -> turn
        // calculator off
        //---------------------------------------
        if (key == 4363 || (TI89 && key == 16651) || (!TI89 && key == 8459)) {
            off();
            OSTimerRestart(APD_TIMER);
        }
        else {
            return key & 0xf7ff;  // mask out 0x800 (repeat flag)
        }
    }
}


//=============================================================================
// rotate a section
//
// the section layout on the screen is as follows:
//
// <section0> <section1>
// <section2> <section3>
//
// section 4 is the center section
//
// direction == 0 ... rotate left
// direction == 1 ... rotate right
//
//=============================================================================
void Rotate(short section,short direction) {
    short a,b,temp;

    //--------------------------------------------------------
    // during the shuffle section numbers larger than 4 may
    // occur (necessary to raise the probability of a center
    // section rotation !!)
    //--------------------------------------------------------
    if (section > 4) section = 4;

    //--------------------------------------------------------
    // setup a & b so that it will hold the coordinates of the
    // left/top corner of the section to rotate
    //--------------------------------------------------------
    switch(section) {
        case 0:
        case 1:
            a = -2 + (section+1)*2;
            b = 0;
            break;
        case 2:
        case 3:
            a = -6 + (section+1)*2;
            b = 2;
            break;
        default:
            a = 1;
            b = 1;
            break;
    }

    temp=board[a][b];

    //-----------------------------------------------------
    // perform the rotation depending on given rotation and
    // set corresponding update flags
    //-----------------------------------------------------
    if (!direction) {
        board[a][b]      = board[a+1][b];
        board[a+1][b]    = board[a+1][b+1];
        board[a+1][b+1]  = board[a][b+1];
        board[a][b+1]    = temp;
        redraw[a][b]     = 1;
        redraw[a+1][b]   = 1;
        redraw[a+1][b+1] = 1;
        redraw[a][b+1]   = 1;
    }
    else {
        board[a][b]      = board[a][b+1];
        board[a][b+1]    = board[a+1][b+1];
        board[a+1][b+1]  = board[a+1][b];
        board[a+1][b]    = temp;
        redraw[a][b]     = 1;
        redraw[a+1][b]   = 1;
        redraw[a+1][b+1] = 1;
        redraw[a][b+1]   = 1;
    }
}


//=============================================================================
// shuffle the board
//
// this is done by performing a specific number of random rotates of random
// sections
//=============================================================================
void Shuffle(void) {
    short i;

    //----------------------------------------------------------
    // copy initial setup to board and set all redraw flags to 1
    //----------------------------------------------------------
    memcpy(board,initial_setup,sizeof(short)*16);
    memset(redraw,1,sizeof(short)*16);

    //------------------------------------------------
    // perform the random rotations of random sections
    //------------------------------------------------
    for (i=0;i<SHUFFLE_LOOPS;i++) {
        //--------------------------------------------
        // NOTE that section numbers larger than 4 are
        // clipped to section 4 (center section)
        //
        // the probability for a center section
        // rotation is higher to make the shuffle more
        // useful
        //--------------------------------------------
        Rotate(random(7),random(2));
    }
}



WIN_RECT marker[6] = {
  {1,1,45,45},
  {47,1,91,45},
  {1,47,45,91},
  {47,47,91,91},
  {24,24,68,68},
  {23,23,69,69}};


//=============================================================================
// redraw frames
//=============================================================================
void RedrawFrames(short oldsection,short newsection) {
    short i;

    if (oldsection != -1) {
        DrawColorRect(marker[oldsection].x0+offsetx,
                      marker[oldsection].y0+offsety,
                      marker[oldsection].x1+offsetx,
                      marker[oldsection].y1+offsety,
                      COLOR_LIGHTGRAY,RECT_EMPTY);
    }

    for (i=0;i<=92;i+=46) {
       DrawColorLine( 1+offsetx, i+offsety,45+offsetx, i+offsety,COLOR_BLACK);
       DrawColorLine(47+offsetx, i+offsety,91+offsetx, i+offsety,COLOR_BLACK);
       DrawColorLine( i+offsetx, 1+offsety, i+offsetx,45+offsety,COLOR_BLACK);
       DrawColorLine( i+offsetx,47+offsety, i+offsetx,91+offsety,COLOR_BLACK);
    }

    for (i=23;i<=69;i+=46) {
       DrawColorLine( 1+offsetx, i+offsety,45+offsetx, i+offsety,COLOR_WHITE);
       DrawColorLine(47+offsetx, i+offsety,91+offsetx, i+offsety,COLOR_WHITE);
       DrawColorLine( i+offsetx, 1+offsety, i+offsetx,45+offsety,COLOR_WHITE);
       DrawColorLine( i+offsetx,47+offsety, i+offsetx,91+offsety,COLOR_WHITE);
    }

    DrawColorRect(marker[newsection].x0+offsetx,
                  marker[newsection].y0+offsety,
                  marker[newsection].x1+offsetx,
                  marker[newsection].y1+offsety,
                  COLOR_BLACK,RECT_EMPTY);

    if (newsection == 4) {
        DrawColorRect(marker[5].x0+offsetx,
                      marker[5].y0+offsety,
                      marker[5].x1+offsetx,
                      marker[5].y1+offsety,
                      COLOR_BLACK,RECT_EMPTY);
    }
}


//=============================================================================
// loop over complete board and redraw a square if the corresponding
// redraw flag is set to 1
//=============================================================================
void DrawBoard(short oldsection,short newsection) {
    short i,j;
    char buffer[2];

    buffer[1] = 0;

    for (i=0;i<4;i++) {
        for (j=0;j<4;j++) {
            if (redraw[i][j]) DrawField(i,j);
        }
    }

    RedrawFrames(oldsection,newsection);
}


//=============================================================================
// initialize a new game
//
// clears the complete screen, shuffles the initial setup and draws the board
//
// returns initial section (4 == center section)
//=============================================================================
short InitGame(void) {
    resettime = 1;

    randomize();
    memset(GetPlane(0),0,240*128/8);
    memset(GetPlane(1),0,240*128/8);
    Shuffle();
    DrawBoard(-1,4);

    DrawString(104,10,"TI-LIDS",F_6x8,A_REPLACE | A_SHADOWED);
    DrawString(110,20,LIDS_VERSION,F_6x8,A_REPLACE | A_SHADOWED);

/*
    DrawString(94,35,"Select",F_4x6,A_REPLACE);
    DrawString(94,43,"Rotate",F_4x6,A_REPLACE);
    DrawString(94,51,"New",F_4x6,A_REPLACE);
    DrawString(94,59,"Exit",F_4x6,A_REPLACE);

    DrawString(118,35,"- Up/Down",F_4x6,A_REPLACE);
    DrawString(118,43,"- Left/Right",F_4x6,A_REPLACE);
    DrawString(118,51,"- Clear",F_4x6,A_REPLACE);
    DrawString(118,59,"- Esc",F_4x6,A_REPLACE);
*/
    DrawString(94,35,"Select",F_4x6,A_REPLACE);
    DrawString(94,51,"Rotate",F_4x6,A_REPLACE);
    DrawString(94,59,"New",F_4x6,A_REPLACE);
    DrawString(94,67,"Exit",F_4x6,A_REPLACE);

    DrawString(118,35,"- Up/Down",F_4x6,A_REPLACE);
    DrawString(118,43,"- 7/9/5/1/3",F_4x6,A_REPLACE);
    DrawString(118,51,"- Left/Right",F_4x6,A_REPLACE);
    DrawString(118,59,"- Clear",F_4x6,A_REPLACE);
    DrawString(118,67,"- Esc",F_4x6,A_REPLACE);

    DrawString(102,77,"(c) TICT 2002",F_4x6,A_REPLACE | A_SHADOWED);

    PrintTime();
    running = 1;

    return 4;
}


//=============================================================================
// Handle winning situation
//
//check is game is won
//=============================================================================
short HandleWon(short section) {
    short i,j,sy;

    if (board[0][0] != board[1][1]) return section;

    for (i=0; i<3; i+=2) {
        for (j=0; j<3; j+=2) {
            if (board[j][i]    != board[j][i+1]   ||
                board[j][i+1]  != board[j+1][i+1] ||
                board[j+1][i+1]!= board[j+1][i])
            {
                return section;
            }
        }
    }

    running = 0;

    sy = LCD_HEIGHT/2 - 2;
    DrawPopup(4,sy-10,LCD_WIDTH-5,sy+10);
    DrawString(0,sy-3,"Perfect. You have won!",F_6x8,A_REPLACE | A_CENTERED | A_SHADOWED);

    GetUserInput();

    return InitGame();
}



#include "logo.c"


//=============================================================================
// as usual our main routine ...
//=============================================================================
void _main(void) {
    short key;
    short section;

    if (!GrayOn()) return;

    if (TI89) offsetx = offsety = 0;
    else      offsetx = (240-160)/2,offsety = (128-100)/2;

    outputtime = mseconds50 = seconds = minutes = hours = 0;

    Logo();

    section = InitGame();

    oldint5 = GetIntVec(AUTO_INT_5);
    SetIntVec(AUTO_INT_5, myint5handler);

    do {
        key = GetUserInput();

        if (key == KEY_LEFT) {
            Rotate(section,0);
            DrawBoard(-1,section);
            section = HandleWon(section);
        }
        else if (key == KEY_RIGHT) {
            Rotate(section,1);
            DrawBoard(-1,section);
            section = HandleWon(section);
        }
        else if (key == KEY_UP) {
            short oldsec = section;
            switch(section) {
                case 0: section = 3; break;
                case 1: section = 0; break;
                case 2: section = 4; break;
                case 3: section = 2; break;
                case 4: section = 1; break;
            }
            DrawBoard(oldsec,section);
        }
        else if (key == KEY_DOWN) {
            short oldsec = section;
            switch(section) {
                case 0: section = 1; break;
                case 1: section = 4; break;
                case 2: section = 3; break;
                case 3: section = 0; break;
                case 4: section = 2; break;
            }
            DrawBoard(oldsec,section);
        }
        else if (key == KEY_SECTION1) {
            short oldsec = section;
            section = 0;
            DrawBoard(oldsec,section);
        }
        else if (key == KEY_SECTION2) {
            short oldsec = section;
            section = 1;
            DrawBoard(oldsec,section);
        }
        else if (key == KEY_SECTION3) {
            short oldsec = section;
            section = 2;
            DrawBoard(oldsec,section);
        }
        else if (key == KEY_SECTION4) {
            short oldsec = section;
            section = 3;
            DrawBoard(oldsec,section);
        }
        else if (key == KEY_SECTION5) {
            short oldsec = section;
            section = 4;
            DrawBoard(oldsec,section);
        }
        else if (key == KEY_CLEAR) {
            section = InitGame();
        }
    }
    while (key != KEY_ESC);

    SetIntVec(AUTO_INT_5, oldint5);

    GrayOff();
}
